import os
from mitosheet.tests.decorators import requires_streamlit
from mitosheet.streamlit.v1.spreadsheet import RunnableAnalysis
import pytest
import pandas as pd

from mitosheet.tests.test_transpile import custom_import


simple_fn = f"""from mitosheet.public.v3 import *
import pandas as pd

def function(import_dataframe_0):
    return import_dataframe_0
"""

simple_param_metadata = [
    {
        'original_value': 'test_df_name',
        'type': 'df_name',
        'subtype': 'import_dataframe',
        'required': True,
        'name': 'import_dataframe_0'
    },
]

@requires_streamlit
def test_mito_analysis_run(tmp_path):
    fully_parameterized_function = f"""from mitosheet.public.v3 import *
import pandas as pd

def function(file_name_import_csv_0, file_name_import_excel_0, file_name_export_csv_0, file_name_export_excel_0):
    txt = pd.read_csv(file_name_import_csv_0)
    
    sheet_df_dictonary = pd.read_excel(file_name_import_excel_0, engine='openpyxl', sheet_name=[
        'Sheet1'
    ], skiprows=0)
    Sheet1 = sheet_df_dictonary['Sheet1']

    txt.insert(1, 'HelloWorld', SUM(1,2))
    
    txt.to_csv(file_name_export_csv_0, index=False)
    
    with pd.ExcelWriter(file_name_export_excel_0, engine="openpyxl") as writer:
        Sheet1.to_excel(writer, sheet_name="Sheet1", index=False)
    
    return txt, Sheet1
"""
    tmp_file1 = str(tmp_path / 'txt.csv')
    tmp_file2 = str(tmp_path / 'file.xlsx')
    tmp_exportfile1 = str(tmp_path / 'export.csv')
    tmp_exportfile2 = str(tmp_path / 'export.xlsx')
    df1 = pd.DataFrame({'A': [1], 'B': [2]})
    expected_df = pd.DataFrame({ 'A': [1], 'HelloWorld': [3], 'B': [2] })
    df1.to_csv(tmp_file1, index=False)
    df1.to_excel(tmp_file2, index=False)
    param_metadata = [
        {
            'original_value': tmp_file1,
            'type': 'import',
            'subtype': 'file_name_import_csv',
            'name': 'file_name_import_csv_0',
            'required': False
        },
        {
            'original_value': tmp_file2,
            'type': 'import',
            'subtype': 'file_name_import_excel',
            'name': 'file_name_import_excel_0',
            'required': False
        },
        {
            'original_value': tmp_exportfile1,
            'type': 'export',
            'subtype': 'file_name_export_csv',
            'name': 'file_name_export_csv_0',
            'required': False
        },
        {
            'original_value': tmp_exportfile2,
            'type': 'export',
            'subtype': 'file_name_export_excel',
            'name': 'file_name_export_excel_0',
            'required': False
        }
    ]
    
    analysis = RunnableAnalysis('', None, fully_parameterized_function, param_metadata)
    result = analysis.run()
    assert result is not None
    pd.testing.assert_frame_equal(result[0], expected_df)
    pd.testing.assert_frame_equal(result[1], df1)

@requires_streamlit
def test_mito_analysis_run_with_args(tmp_path):
    fully_parameterized_function = f"""from mitosheet.public.v3 import *
import pandas as pd

def function(file_name_import_csv_0, file_name_import_excel_0, file_name_export_csv_0, file_name_export_excel_0):
    txt = pd.read_csv(file_name_import_csv_0)
    
    sheet_df_dictonary = pd.read_excel(file_name_import_excel_0, engine='openpyxl', sheet_name=[
        'Sheet1'
    ], skiprows=0)
    Sheet1 = sheet_df_dictonary['Sheet1']

    txt.insert(1, 'HelloWorld', SUM(1,2))
    
    txt.to_csv(file_name_export_csv_0, index=False)
    
    with pd.ExcelWriter(file_name_export_excel_0, engine="openpyxl") as writer:
        Sheet1.to_excel(writer, sheet_name="Sheet1", index=False)
    
    return txt, Sheet1
"""
    tmp_file1 = str(tmp_path / 'txt.csv')
    tmp_file2 = str(tmp_path / 'file.xlsx')
    tmp_exportfile1 = str(tmp_path / 'export.csv')
    tmp_exportfile2 = str(tmp_path / 'export.xlsx')
    df1 = pd.DataFrame({'A': [1], 'B': [2]})
    expected_df = pd.DataFrame({ 'A': [1], 'HelloWorld': [3], 'B': [2] })
    df1.to_csv(tmp_file1, index=False)
    df1.to_excel(tmp_file2, index=False)
    param_metadata = [
        {
            'original_value': tmp_file1,
            'type': 'import',
            'subtype': 'file_name_import_csv',
            'name': 'file_name_import_csv_0',
            'required': False
        },
        {
            'original_value': tmp_file2,
            'type': 'import',
            'subtype': 'file_name_import_excel',
            'name': 'file_name_import_excel_0',
            'required': False
        },
        {
            'original_value': tmp_exportfile1,
            'type': 'export',
            'subtype': 'file_name_export_csv',
            'name': 'file_name_export_csv_0',
            'required': False
        },
        {
            'original_value': tmp_exportfile2,
            'type': 'export',
            'subtype': 'file_name_export_excel',
            'name': 'file_name_export_excel_0',
            'required': False
        }
    ]
    
    new_export_file_0 = str(tmp_path / 'new_export.csv')
    new_export_file_1 = str(tmp_path / 'new_export.xlsx')
    analysis = RunnableAnalysis('', None, fully_parameterized_function, param_metadata)

    # Test that get_param_metadata works
    assert analysis.get_param_metadata() == param_metadata
    assert analysis.get_param_metadata(param_type='import') == param_metadata[:2]
    assert analysis.get_param_metadata(param_type='export') == param_metadata[2:]

    with pytest.raises(TypeError):
        analysis.get_param_metadata(param_type='invalid')

    result = analysis.run(file_name_export_csv_0=new_export_file_0)
    assert result is not None
    assert os.path.exists(new_export_file_0)
    assert not os.path.exists(new_export_file_1)
    pd.testing.assert_frame_equal(result[0], expected_df)
    pd.testing.assert_frame_equal(result[1], df1)


@requires_streamlit
def test_mito_analysis_run_with_import_var_in_code(tmp_path):
    fully_parameterized_function = f"""from mitosheet.public.v3 import *

def function_ctqm(import_dataframe_0):
    import_dataframe_0.insert(1, 'Test', 0)
    
    return import_dataframe_0
"""
    df = pd.DataFrame({'A': [1], 'B': [2]})
    param_metadata = [
        {
            'original_value': 'test_df_name',
            'type': 'import',
            'subtype': 'import_dataframe',
            'required': True,
            'name': 'import_dataframe_0'
        }
    ]

    analysis = RunnableAnalysis('', None, fully_parameterized_function, param_metadata)

    assert analysis.get_param_metadata() == param_metadata
    assert analysis.get_param_metadata('import') == param_metadata
    assert analysis.get_param_metadata('export') == []

    result = analysis.run(df)
    expected_df = pd.DataFrame({'A': [1], 'Test': [0], 'B': [2]})
    assert result is not None
    pd.testing.assert_frame_equal(result, expected_df)
    
@requires_streamlit
def test_mito_analysis_run_with_positional_args(tmp_path):
    fully_parameterized_function = f"""from mitosheet.public.v3 import *
import pandas as pd

def function(import_dataframe_0, file_name_import_csv_0, file_name_import_excel_0, file_name_export_csv_0, file_name_export_excel_0):
    txt = pd.read_csv(file_name_import_csv_0)
    
    sheet_df_dictonary = pd.read_excel(file_name_import_excel_0, engine='openpyxl', sheet_name=[
        'Sheet1'
    ], skiprows=0)
    Sheet1 = sheet_df_dictonary['Sheet1']

    txt.insert(1, 'HelloWorld', SUM(1,2))
    
    txt.to_csv(file_name_export_csv_0, index=False)
    
    with pd.ExcelWriter(file_name_export_excel_0, engine="openpyxl") as writer:
        Sheet1.to_excel(writer, sheet_name="Sheet1", index=False)
    
    return import_dataframe_0, txt, Sheet1
"""
    tmp_file1 = str(tmp_path / 'txt.csv')
    tmp_file2 = str(tmp_path / 'file.xlsx')
    tmp_exportfile1 = str(tmp_path / 'export.csv')
    tmp_exportfile2 = str(tmp_path / 'export.xlsx')
    df1 = pd.DataFrame({'A': [1], 'B': [2]})
    expected_df = pd.DataFrame({ 'A': [1], 'HelloWorld': [3], 'B': [2] })
    df1.to_csv(tmp_file1, index=False)
    df1.to_excel(tmp_file2, index=False)
    param_metadata = [
        {
            'original_value': 'test_df_name',
            'type': 'import',
            'subtype': 'import_dataframe',
            'required': True,
            'name': 'import_dataframe_0'
        },
        {
            'original_value': tmp_file1,
            'type': 'import',
            'subtype': 'file_name_import_csv',
            'name': 'file_name_import_csv_0',
            'required': False
        },
        {
            'original_value': tmp_file2,
            'type': 'import',
            'subtype': 'file_name_import_excel',
            'name': 'file_name_import_excel_0',
            'required': False
        },
        {
            'original_value': tmp_exportfile1,
            'type': 'export',
            'subtype': 'file_name_export_csv',
            'name': 'file_name_export_csv_0',
            'required': False
        },
        {
            'original_value': tmp_exportfile2,
            'type': 'export',
            'subtype': 'file_name_export_excel',
            'name': 'file_name_export_excel_0',
            'required': False
        }
    ]
    
    new_export_file_0 = str(tmp_path / 'new_export.csv')
    new_export_file_1 = str(tmp_path / 'new_export.xlsx')
    analysis = RunnableAnalysis('', None, fully_parameterized_function, param_metadata)

    # Test that get_param_metadata works
    assert analysis.get_param_metadata() == param_metadata
    assert analysis.get_param_metadata(param_type='import') == param_metadata[:3]
    assert analysis.get_param_metadata(param_type='export') == param_metadata[3:]

    new_df = pd.DataFrame({'A': [1], 'B': [2]})
    result = analysis.run(df1, file_name_export_csv_0=new_export_file_0)
    assert result is not None
    assert os.path.exists(new_export_file_0)
    assert not os.path.exists(new_export_file_1)
    pd.testing.assert_frame_equal(result[0], df1)
    pd.testing.assert_frame_equal(result[1], expected_df)
    pd.testing.assert_frame_equal(result[2], new_df)


@requires_streamlit
def test_mito_analysis_run_with_duplicate_positional_and_kwargs(tmp_path):
    fully_parameterized_function = f"""from mitosheet.public.v3 import *
import pandas as pd

def function(import_dataframe_0, file_name_import_csv_0, file_name_import_excel_0, file_name_export_csv_0, file_name_export_excel_0):
    txt = pd.read_csv(file_name_import_csv_0)
    
    sheet_df_dictonary = pd.read_excel(file_name_import_excel_0, engine='openpyxl', sheet_name=[
        'Sheet1'
    ], skiprows=0)
    Sheet1 = sheet_df_dictonary['Sheet1']

    txt.insert(1, 'HelloWorld', SUM(1,2))
    
    txt.to_csv(file_name_export_csv_0, index=False)
    
    with pd.ExcelWriter(file_name_export_excel_0, engine="openpyxl") as writer:
        Sheet1.to_excel(writer, sheet_name="Sheet1", index=False)
    
    return import_dataframe_0, txt, Sheet1
"""
    tmp_file1 = str(tmp_path / 'txt.csv')
    tmp_file2 = str(tmp_path / 'file.xlsx')
    tmp_exportfile1 = str(tmp_path / 'export.csv')
    tmp_exportfile2 = str(tmp_path / 'export.xlsx')
    df1 = pd.DataFrame({'A': [1], 'B': [2]})
    expected_df = pd.DataFrame({ 'A': [1], 'HelloWorld': [3], 'B': [2] })
    df1.to_csv(tmp_file1, index=False)
    df1.to_excel(tmp_file2, index=False)
    param_metadata = [
        {
            'original_value': 'test_df_name',
            'type': 'import',
            'subtype': 'import_dataframe',
            'required': True,
            'name': 'import_dataframe_0'
        },
        {
            'original_value': tmp_file1,
            'type': 'import',
            'subtype': 'file_name_import_csv',
            'name': 'file_name_import_csv_0',
            'required': False
        },
        {
            'original_value': tmp_file2,
            'type': 'import',
            'subtype': 'file_name_import_excel',
            'name': 'file_name_import_excel_0',
            'required': False
        },
        {
            'original_value': tmp_exportfile1,
            'type': 'export',
            'subtype': 'file_name_export_csv',
            'name': 'file_name_export_csv_0',
            'required': False
        },
        {
            'original_value': tmp_exportfile2,
            'type': 'export',
            'subtype': 'file_name_export_excel',
            'name': 'file_name_export_excel_0',
            'required': False
        }
    ]
    
    new_export_file_0 = str(tmp_path / 'new_export.csv')
    analysis = RunnableAnalysis('', None, fully_parameterized_function, param_metadata)
    with pytest.raises(TypeError):
        analysis.run(df1, import_dataframe_0=df1, file_name_export_csv_0=new_export_file_0)
    


@requires_streamlit
def test_run_fail_missing_required():
    fully_parameterized_function = f"""from mitosheet.public.v3 import *
import pandas as pd

def function(import_dataframe_0):
    return import_dataframe_0
"""
    param_metadata = [
        {
            'original_value': 'test_df_name',
            'type': 'import',
            'subtype': 'import_dataframe',
            'required': True,
            'name': 'import_dataframe_0'
        },
    ]
    
    analysis = RunnableAnalysis('', None, fully_parameterized_function, param_metadata)
    with pytest.raises(TypeError):
        analysis.run()
    

@requires_streamlit
def test_run_fail_incorrect_args():
    fully_parameterized_function = f"""from mitosheet.public.v3 import *
import pandas as pd

def function(import_dataframe_0):
    return import_dataframe_0
"""
    param_metadata = [
        {
            'original_value': "xyz",
            'type': 'export',
            'subtype': 'file_name_export_excel',
            'name': 'file_name_export_excel_0',
            'required': False
        }
    ]
    
    analysis = RunnableAnalysis('', None, fully_parameterized_function, param_metadata)
    with pytest.raises(TypeError):
        analysis.run(testing=1)
    

@requires_streamlit
def test_to_and_from_json():
    analysis = RunnableAnalysis('', None, simple_fn, simple_param_metadata)
    # Test that the to_json function 
    json = analysis.to_json()
    assert json is not None
    assert json == r'{"code": "", "code_options": null, "fully_parameterized_function": "from mitosheet.public.v3 import *\nimport pandas as pd\n\ndef function(import_dataframe_0):\n    return import_dataframe_0\n", "param_metadata": [{"original_value": "test_df_name", "type": "df_name", "subtype": "import_dataframe", "required": true, "name": "import_dataframe_0"}], "mito_analysis_version": 1}'

    # Test that the from_json function works
    new_analysis = RunnableAnalysis.from_json(json)
    assert new_analysis is not None
    
    df = pd.DataFrame({'A': [1], 'B': [2]})
    result = new_analysis.run(df)
    assert result is not None
    pd.testing.assert_frame_equal(result, df)

@requires_streamlit
def test_to_and_from_json_special_characters():
    special_characters_fn = """from mitosheet.public.v3 import *
import pandas as pd

def function(vari\abl"e_name{}):
    return vari\abl"e_name{}
"""
    special_characters_metadata = [
    {
        'original_value': 'test_df_name',
        'type': 'df_name',
        'subtype': 'import_dataframe',
        'required': True,
        'name': 'vari\ abl"e_name{}'
    },
]
    analysis = RunnableAnalysis('', None, special_characters_fn, special_characters_metadata)
    # Test that the to_json function 
    json = analysis.to_json()
    assert json is not None

    # Test that the from_json function works
    new_analysis = RunnableAnalysis.from_json(json)
    assert new_analysis is not None
    assert new_analysis.fully_parameterized_function == special_characters_fn

@requires_streamlit
def test_custom_imports():
    fully_parameterized_code = """from mitosheet.public.v3 import *
from mitosheet.tests.test_transpile import custom_import, ADDONE

def function():
    df1 = custom_import()
    
    df1.insert(1, 'B', ADDONE(df1['A']))
    
    return df1
"""
    analysis = RunnableAnalysis('', None, fully_parameterized_code, [])
    result = analysis.run()
    pd.testing.assert_frame_equal(result, pd.DataFrame({'A': [1, 2, 3], 'B': [2, 3, 4]}))

@requires_streamlit
def test_can_pass_dataframe_to_file_path_in_run_auto_conversion():

    analysis = RunnableAnalysis(
        '',
        None,
        """from mitosheet.public.v3 import *
import pandas as pd

def function_srjr(file_name_import_csv_0):
    df_export = pd.read_csv(file_name_import_csv_0)
    
    return df_export
""",
    [{"original_value": "datasets/df_export.csv", "type": "import", "subtype": "file_name_import_csv", "required": False, "name": "file_name_import_csv_0"}]
)

    # Run with a df rather than a file path
    df = pd.DataFrame({'A': [1, 2, 3]})
    result = analysis.run(file_name_import_csv_0=df)
    assert result.equals(df)


#!/usr/bin/env python
# coding: utf-8

# Copyright (c) Saga Inc.
# Distributed under the terms of the GPL License.
import base64
import io
from typing import Any, Dict

from mitosheet.types import StepsManagerType
from mitosheet.user import is_pro
from mitosheet.user.utils import is_running_test
from mitosheet.utils import write_to_excel


def get_dataframe_as_excel(params: Dict[str, Any], steps_manager: StepsManagerType) -> str:
    """
    Sends a dataframe as a excel string.
    """
    sheet_indexes = params['sheet_indexes']

    # Formatting is a Mito pro feature, but we also allow it for testing
    export_formatting = params.get('export_formatting', False)
    allow_formatting = (is_pro() or is_running_test()) and export_formatting

    # We write to a buffer so that we don't have to save the file
    # to the file system for no reason
    buffer = io.BytesIO()
    write_to_excel(buffer, sheet_indexes, steps_manager.curr_step.post_state, allow_formatting=allow_formatting)    
    # Go back to the start of the buffer
    buffer.seek(0)
    
    # First, we take the buffer, and base64 encode it in bytes,
    # and then we covert this to ASCII. On the front-end, we 
    # turn it back into base64, then back to bytes, before 
    # creating a Blob out of it
    return base64.b64encode(buffer.read()).decode('ascii')



from dash import Dash, callback, Input, Output, html, dcc
from mitosheet.mito_dash.v1 import Spreadsheet, mito_callback, activate_mito

app = Dash(__name__)
activate_mito(app)

CSV_URL = 'https://raw.githubusercontent.com/plotly/datasets/master/tesla-stock-price.csv'

app.layout = html.Div([
    html.H1("Stock Analysis"),
    Spreadsheet(CSV_URL, id={'type': 'spreadsheet', 'id': 'sheet'}),
    html.Div(id='output')
])

@mito_callback(
    Output('output', 'children'),
    Input({'type': 'spreadsheet', 'id': 'sheet'}, 'spreadsheet_result'),
)
def update_code(spreadsheet_result):
    return html.Div([
        html.Code(spreadsheet_result.code())
    ])
    
if __name__ == '__main__':
    app.run_server(debug=True)


from mitosheet.mito_dash.v1 import Spreadsheet, mito_callback, activate_mito
from dash import Dash, html, Input, Output
import pandas as pd

app = Dash(__name__)
activate_mito(app)

df = pd.read_csv('https://raw.githubusercontent.com/plotly/datasets/master/gapminder2007.csv')

app.layout = html.Div([
    html.H1('Simple Dataframe Editing Example'),
    Spreadsheet(df, id={'type': 'spreadsheet', 'id': 'sheet'}),
    html.Div(id='output')
])

@mito_callback(
    Output('output', 'children'),
    Input({'type': 'spreadsheet', 'id': 'sheet'}, 'spreadsheet_result'),
)
def update_output(spreadsheet_result):
    return html.Div([
        html.H3('Edited Dataframes'),
        html.Div(f'Dataframes: {spreadsheet_result.dfs()}')
    ])

if __name__ == '__main__':
    app.run_server(debug=True)


#!/usr/bin/env python
# coding: utf-8

# Copyright (c) Saga Inc.
# Distributed under the terms of the GPL License.
from abc import ABC, abstractmethod
from time import perf_counter
from typing import Any, Dict, List, Optional, Set, Tuple

from mitosheet.code_chunks.code_chunk import CodeChunk
from mitosheet.state import State
from mitosheet.transpiler.transpile_utils import get_globals_for_exec
from mitosheet.types import (ColumnHeader, ColumnID,
                             ExecuteThroughTranspileNewDataframeParams, StepType)



class StepPerformer(ABC, object):
    """
    The abstract base class for a step performer, which are the set
    of classes that execute, transpile, and describe steps. 
    """

    @classmethod
    @abstractmethod
    def step_version(cls) -> int:
        """
        Returns the version of the step. Changes when the parameters
        of the step change.
        """
        pass

    @classmethod
    @abstractmethod
    def step_type(cls) -> str:
        """
        The name of the step used internally. If you change this, you must upgrade
        the step and bump the version.
        """
        pass
    
    @classmethod
    def step_event_type(cls) -> str:
        """
        The type of the edit event that generates this type of step. 
        This is always just f'{cls.step_type}_edit'.
        """
        return f'{cls.step_type()}_edit'

    @classmethod
    def saturate(cls, prev_state: State, params: Dict[str, Any], previous_steps: List[StepType]) -> Dict[str, Any]:
        """
        Given the parameters of the step, will saturate the event with
        more parameters based on the passed prev_state. 
        """
        # By default, we don't do anything with the saturate
        return params

    @classmethod
    def execute(cls, prev_state: State, params: Dict[str, Any]) -> Tuple[State, Optional[Dict[str, Any]]]:
        """
        1. Parse params
        2. Error check (using no functions specific to Pandas)
        3. Build necessary execution data - namely, new column header info or new dataframe info
        3. Execute the transpiled code to make the new state
        4. Build a result from this

        Execute always returns the post_state, and optionally returns a dictionary
        of execution_data, which is data that may be useful to the transpiler in
        transpiling the code.

        If the execution_data also includes the key `pandas_processing_time`, this 
        will allow the logging infrastructure to determine how much overhead Mito
        adds to executing this event.

        By default, this function will return the result of just executing the transpiled
        code. 

        In general, this function should use _no_ Pandas specific functionality, but rather 
        should just be a) updating metadata or b) computing results from executed code
        that is useful to the spreadsheet.
        """
        post_state, execution_data = cls.execute_through_transpile(prev_state, params)
        return post_state, execution_data

    @classmethod
    @abstractmethod
    def transpile(
        cls,
        prev_state: State,
        params: Dict[str, Any],
        execution_data: Optional[Dict[str, Any]],
    ) -> List[CodeChunk]:
        """
        Returns a list of the CodeChunks that correspond to this 
        step being executed
        """
        pass

    @classmethod
    def execute_through_transpile(
        cls,
        prev_state: State,
        params: Dict[str, Any],
        execution_data: Optional[Dict[str, Any]]=None,
        new_dataframe_params: Optional[ExecuteThroughTranspileNewDataframeParams]=None,
        column_headers_to_column_ids: Optional[Dict[ColumnHeader, ColumnID]]=None,
        optional_code: Optional[Tuple[List[str], List[str]]]=None,
        use_deprecated_id_algorithm: bool=False
    ) -> Tuple[State, Dict[str, Any]]:
        """
        Previously, we had a ton of duplicated code in the execute and the code chunk functions. This was annoying 
        for a few reasons:
        1. Duplicated code introduced the possibility that things worked differently on our backend vs. in our generated code. 
        2. It was more work to write, since you had to write everything twice.
        3. It meant we were tied to a specific programming framework like pandas 

        This function was introduced so that we can instead just do the `execute` method simply by executing the 
        code that we would have generated for this step. Thus, the only thing `execute` ends up being responsible 
        for is doing error handling, updating some metadata about the step, or computing a result to return to the
        frontend. In other words: everything other than the actual execution, which is handled by this code here.

        This code works similarly to the AI recon, in that it uses an exec statement and then looks at what changed 
        between before and after the code is executed. However, this code has the additional benefit of knowing what
        it expects to change, so the various parameters are how the caller of this method can tell this function what
        should be different before and after in dataframes.
        """
        from mitosheet.ai.recon import update_state_by_reconing_dataframes

        if execution_data is None:
            execution_data = {}

        modified_dataframe_indexes = cls.get_modified_dataframe_indexes(params)
        # If the modified indexes are -1, then only new dataframes have been created -- and in this
        # case we just don't detect modifications
        if modified_dataframe_indexes == {-1}:
            modified_dataframe_indexes = set()

        post_state = prev_state.copy(deep_sheet_indexes=modified_dataframe_indexes)

        code_chunks = cls.transpile(post_state, params, execution_data)
        code = []
        for chunk in code_chunks:
            _code, imports = chunk.get_code()
            code.extend(imports)
            code.extend(_code)

        final_code = "\n".join(code)

        # TODO: this is weird. This will not always be updated, accoring to exec documentation, 
        # but in practice is seems to work...
        exec_globals = get_globals_for_exec(post_state, post_state.public_interface_version)
        exec_locals = {**exec_globals}
        
        pandas_start_time = perf_counter()
        exec(final_code, exec_globals, exec_locals)

        # Go through the optional code lines
        optional_code_that_successfully_executed: Tuple[List[str], List[str]] = ([], [])
        if optional_code is not None:
            for optional_import in optional_code[1]:
                try:
                    exec(optional_import, exec_globals, exec_locals)
                    optional_code_that_successfully_executed = (
                        optional_code_that_successfully_executed[0],
                        optional_code_that_successfully_executed[1] + [optional_import],
                    )
                except:
                    break
            
            # Take special care to not add the comments for the code unless 
            # the code itself executes successfully

            non_code_lines_before_optional_line = []
            for optional_code_line in optional_code[0]:

                # We don't need to exec spaces
                if optional_code_line == '' or optional_code_line.strip().startswith('#'):
                    non_code_lines_before_optional_line.append(optional_code_line)
                    continue

                # TODO: we should make it so it rolls back the state if this fails
                # but it's fine for now -- since partial updates don't seem to 
                # manifest in practice
                try:
                    exec(optional_code_line, exec_globals, exec_locals)
                    optional_code_that_successfully_executed = (
                        optional_code_that_successfully_executed[0] + non_code_lines_before_optional_line + [optional_code_line],
                        optional_code_that_successfully_executed[1],
                    )
                    non_code_lines_before_optional_line = []
                except:
                    break


        pandas_processing_time = perf_counter() - pandas_start_time

        for modified_dataframe_index in modified_dataframe_indexes:
            df_name = prev_state.df_names[modified_dataframe_index]
            new_df = exec_locals[df_name]
            post_state, _ = update_state_by_reconing_dataframes(
                post_state, 
                modified_dataframe_index, 
                prev_state.dfs[modified_dataframe_index],
                new_df, 
                column_headers_to_column_ids=column_headers_to_column_ids
            )

        if new_dataframe_params:
            for new_df_name in new_dataframe_params['new_df_names']:
                df_source = new_dataframe_params['df_source']

                new_df = exec_locals[new_df_name] 
                post_state.add_df_to_state(
                    new_df, 
                    df_name=new_df_name, 
                    df_source=df_source, 
                    overwrite=new_dataframe_params['overwrite'], 
                    use_deprecated_id_algorithm=use_deprecated_id_algorithm
                )

        return post_state, {
            'pandas_processing_time': pandas_processing_time,
            'optional_code_that_successfully_executed': optional_code_that_successfully_executed,
            **execution_data
        }
        

    @classmethod
    @abstractmethod
    def get_modified_dataframe_indexes(cls, params: Dict[str, Any]) -> Set[int]:
        """
        Returns a set of all the sheet indexes that were modified
        by this step.

        If it returns an empty set, then this step
        modified every dataframe.

        If it returned -1, then it modified all new dataframes (on
        the left side of the dfs array).
        """
        pass


from mitosheet.mito_dash.v1 import Spreadsheet, activate_mito
from dash import Dash, html
import pandas as pd

app = Dash(__name__)

try:
    Spreadsheet(id={'type': 'spreadsheet', 'id': 'sheet'})
    assert False
except Exception as e:
    not_activate_mito = e

activate_mito(app)

try:
    Spreadsheet(id='sheet') # type: ignore
    assert False
except Exception as e:
    incorrect_id_error = e


app.layout = html.Div([
    html.H1('Errors were correctly generated!'),
    html.Div("Error for not activated: " + str(not_activate_mito)),
    html.Div("Error for invalid id type: " + str(incorrect_id_error))
])

if __name__ == '__main__':
    app.run_server(debug=True)


#!/usr/bin/env python
# coding: utf-8

# Copyright (c) Jupyter Development Team.
# Distributed under the terms of the Modified BSD License.

"""
The mitosheet package is distributed under the mitosheet and mitosheet3
package names on pip. The package.json will tell you all you need to 
know about which one we are in currently using.

As such, this setup.py script reads in the package.json and sets up
the proper package.
"""

import json
import os
from glob import glob
from pathlib import Path
from typing import List, Tuple

from setuptools import find_packages, setup

HERE = Path(__file__).parent.resolve()

package_json = json.loads(open('package.json').read())
lab_path = Path(HERE, 'mitosheet', 'labextension')
notebook_path = Path(HERE, 'mitosheet', 'nbextension')

data_files_spec = [
    # Notebook extension data files
    ('share/jupyter/nbextensions/mitosheet', notebook_path, '**'),
    ('etc/jupyter/nbconfig/notebook.d', str(HERE), 'mitosheet.json'),

    # Lab extension data files
    ("share/jupyter/labextensions/mitosheet", str(lab_path), "**"),
    ("share/jupyter/labextensions/mitosheet", str(HERE), "install.json"),
]


def get_data_files_from_data_files_spec(
    data_specs: List[Tuple[str, str, str]],
):
    """
    Given tuples of (data_file_path, directory_to_search, pattern_to_find),
    this function will return a list of tuples of (data_file_path, [files])
    in the format that setuptools expects.
    """

    file_data = {}

    for (data_file_path, directory_to_search, pattern_to_find) in data_specs or []:

        # Get the directory to search ready
        if os.path.isabs(directory_to_search):
            directory_to_search = os.path.relpath(directory_to_search)
        directory_to_search = directory_to_search.rstrip("/")

        # Get all non-directory files that match the pattern, searching recursively
        files = [
            f for f in glob(
                Path().joinpath(directory_to_search, pattern_to_find).as_posix(), 
                recursive=True
            ) if not os.path.isdir(f)
        ]

        offset = len(directory_to_search) + 1

        for file_name in files:
            relative_path = str(Path(file_name).parent)[offset:]
            full_data_file_path = Path().joinpath(data_file_path, relative_path).as_posix()

            if full_data_file_path not in file_data:
                file_data[full_data_file_path] = []

            file_data[full_data_file_path].append(file_name)

    # Turn to list and sort by length, to be consistent (and maybe cuz we need to for folder creation?)
    data_files = sorted(file_data.items(), key=lambda x: len(x[0]))
    
    return data_files

data_files = get_data_files_from_data_files_spec(data_files_spec)   

setup_args = dict(
    name                    = 'mitosheet',
    version                 = package_json["version"],
    url                     = package_json["homepage"],
    author                  = package_json["author"]["name"],
    author_email            = package_json["author"]["email"],
    description             = package_json["description"],
    license                 = "GNU Affero General Public License v3",
    long_description="""
    To learn more about Mito, checkout out our documentation: https://docs.trymito.io/getting-started/installing-mito\n\n
    Before installing Mito \n\n
    1. Check that you have Python 3.6 or above. To check your version of Python, open a new terminal, and type python3 --version. If you need to install or update Python, restart your terminal after doing so.\n\n
    2. Checkout our terms of service and privacy policy. By installing Mito, you're agreeing to both of them. Please contact us at aaron [@] sagacollab [dot] com with any questions.\n\n
    Installation Instructions \n\n
    For more detailed installation instructions, see our documentation: https://docs.trymito.io/getting-started/installing-mito\n\n
    1. pip install mitosheet\n\n
    2. Launch JupyterLab 3.0 and open a new notebook\n\n
    3. In the notebook, run the following code:\n\n
    import mitosheet\n\n
    mitosheet.sheet()\n\n
    """,
    long_description_content_type = "text/markdown",
    packages                 = find_packages(exclude=['deployment']),
    include_package_data     = True,
    package_data             = {'': ['*.js', '*.css', '*.html']},
    data_files               = data_files,
    install_requires=[        
        "jupyterlab~=3.0",
        # We allow users to have many versions of pandas installed. All functionality should
        # work, with the exception of Excel import, which might require additonal dependencies
        'pandas>=0.24.2',
        'analytics-python',
        # Graphing libraries
        'plotly>=4.14.3',
        'chardet>=3.0.4',
        # For XLSX, reading - we don't fix so works on all python versions
        'openpyxl',
        # xlsxwriter is needed for adding formatting to exported Excel sheets. 
        # We pin to a pretty old version because the formatting functionality hasn't changed in a long time.
        'xlsxwriter>=0.6.9,<=3.1.3',
    ],
    extras_require = {
        'test': [
            'pytest',
            'flake8',
            'types-chardet',
            'types-requests',
            'mypy',
            'pytest_httpserver',
        ],
        'deploy': [
            'wheel', 
            'twine',
            "setuptools==56.0.0"
        ],
        'streamlit': [
            'streamlit>=1.24',
        ],
        'optional_feature_dependencies': [
            # According to this documentation (https://github.com/snowflakedb/snowflake-connector-python),
            # snowflake-connect-python requires at least Python 3.7
            'snowflake-connector-python[pandas]; python_version>="3.7"',
            'streamlit>=1.24',
            'dash>=2.9'
        ]
    },
    zip_safe                = False,
    python_requires         = ">=3.6",
    platforms               = "Linux, Mac OS X, Windows",
    keywords                = ["Jupyter", "JupyterLab", "JupyterLab3"],
    classifiers             = [
        "License :: OSI Approved :: GNU Affero General Public License v3",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Programming Language :: Python :: 3.11",
        "Framework :: Jupyter",
    ],
)

if __name__ == '__main__':
    setup(**setup_args)



from datetime import datetime, timedelta
from typing import Optional, Union

from mitosheet.public.v3.types.float import cast_string_to_float


def cast_to_number(unknown: Union[str, int, float, bool, datetime, timedelta]) -> Optional[Union[int, float]]:
    """
    A super-type of int and float, this number type will keep ints as ints and floats
    as floats. It will cast strings to floats by default.
    """
    if isinstance(unknown, bool):
        return int(unknown)
    elif isinstance(unknown, str):
        return cast_string_to_float(unknown)
    elif isinstance(unknown, int):
        return unknown
    elif isinstance(unknown, float):
        return unknown
    

    return None


#!/usr/bin/env python
# coding: utf-8

# Copyright (c) Saga Inc.
# Distributed under the terms of the GPL License.
"""
For going to a number series.
"""
from typing import Any
from mitosheet.public.v1.sheet_functions.types.to_float_series import to_float_series

import numpy as np
import pandas as pd


def to_int_series(
        unknown_object: Any,
        on_uncastable_arg_element: Any=('default', np.NaN), # Union[Literal['error'], Tuple[Literal['default'], any]]
    ) -> pd.Series:

    float_series = to_float_series(unknown_object, on_uncastable_arg_element=on_uncastable_arg_element)
    # TODO: do we need to handle nan values?
    return float_series.astype(int)



#!/usr/bin/env python
# coding: utf-8

# Copyright (c) Saga Inc.
# Distributed under the terms of the GPL License.

"""
Contains some types that are useful in the Mitosheet package. 

We use type aliases to make many parts of the codebase more
explicit and clear, and make sure to test the types in our 
continous integration
"""

import pandas as pd
import numpy as np
from typing import TYPE_CHECKING, Dict, List, Optional, Union, Tuple, Any
from collections import OrderedDict

GraphID = str
ColumnID = str

# A column header is either a primative type
PrimativeColumnHeader = Union[int, float, bool, str, Optional[str]]
MultiLevelColumnHeader = Union[Tuple[PrimativeColumnHeader, ...], List[PrimativeColumnHeader]]
# To a tuple of primative types (TODO: does this nest further?).
ColumnHeader = Union[PrimativeColumnHeader, MultiLevelColumnHeader]
IndexLabel = Any

# To resolve circular dependencies, we create a StepsManagerType here
if TYPE_CHECKING:
    from mitosheet.steps_manager import StepsManager
    StepsManagerType = StepsManager
    from mitosheet.mito_backend import MitoBackend
    MitoWidgetType = MitoBackend
    from mitosheet.state import State
    StateType = State
    from mitosheet.step import Step
    StepType = Step
else:
    StepsManagerType = Any
    MitoWidgetType = Any
    StateType = Any
    StepType = Any

IndexType = Union[str, int, bool, float]


DataframeFormat = Dict[str, Any]
ColumnFormat = Dict[str, Any]


ConditionalFormatUUID = str

ConditionalFormat = Dict[str, Any] 
"""
ConditionalFormat: {
    format_uuid: string, // Should be a random string!
    columnIDs: ColumnID[],
    filters: FilterType[],
    invalidFilterColumnIDs: ColumnID[]
    color: string | undefined
    backgroundColor: string | undefined
}
"""

ConditionalFormattingInvalidResults = Dict[ConditionalFormatUUID, List[ColumnID]]
ConditionalFormattingCellResults = Dict[ColumnID, Dict[IndexType, Dict[str, Optional[str]]]]

ConditionalFormattingResult = Dict[str, Union[
        ConditionalFormattingInvalidResults, # A list of the invalid columns for a specific filter
        ConditionalFormattingCellResults # The actual formatting results
    ]
]

PivotColumnTransformation = str

RowOffset = int
ParserMatchSubstringRange = Tuple[int, int] # start, end

# If the user does not have the snowflake.connector python package installed,
# we take extra care to make sure that our mypy typing will still pass even though
# that code is not accessible to the user.
try:
    from snowflake.connector import SnowflakeConnection 
except ImportError:
    # This is prety hacky and causes the mypy tests to fail if we get 
    # rid of the #type: ignore, but I'm unable to create a type based on the 
    # SnowflakeConnection connection otherwise because its conditional 
    # on whether the package is installed or not.
    SnowflakeConnection = Any #type: ignore

MitoSafeSnowflakeConnection = Optional[SnowflakeConnection]

FORMULA_ENTIRE_COLUMN_TYPE = 'entire_column'
FORMULA_SPECIFIC_INDEX_LABELS_TYPE = 'specific_index_labels'

# The constants used in the filter step itself as filter conditions
# NOTE: these must be unique (e.g. no repeating names for different types)
FC_EMPTY = "empty"
FC_NOT_EMPTY = "not_empty"
FC_LEAST_FREQUENT = "least_frequent"
FC_MOST_FREQUENT = "most_frequent"

FC_BOOLEAN_IS_TRUE = "boolean_is_true"
FC_BOOLEAN_IS_FALSE = "boolean_is_false"

FC_STRING_CONTAINS = "contains"
FC_STRING_DOES_NOT_CONTAIN = "string_does_not_contain"
FC_STRING_EXACTLY = "string_exactly"
FC_STRING_NOT_EXACTLY = "string_not_exactly"
FC_STRING_STARTS_WITH = "string_starts_with"
FC_STRING_ENDS_WITH = "string_ends_with"
FC_STRING_CONTAINS_CASE_INSENSITIVE = "string_contains_case_insensitive"

FC_NUMBER_EXACTLY = "number_exactly"
FC_NUMBER_NOT_EXACTLY = "number_not_exactly"
FC_NUMBER_GREATER = "greater"
FC_NUMBER_GREATER_THAN_OR_EQUAL = "greater_than_or_equal"
FC_NUMBER_LESS = "less"
FC_NUMBER_LESS_THAN_OR_EQUAL = "less_than_or_equal"
FC_NUMBER_LOWEST = 'number_lowest'
FC_NUMBER_HIGHEST = 'number_highest'

FC_DATETIME_EXACTLY = "datetime_exactly"
FC_DATETIME_NOT_EXACTLY = "datetime_not_exactly"
FC_DATETIME_GREATER = "datetime_greater"
FC_DATETIME_GREATER_THAN_OR_EQUAL = "datetime_greater_than_or_equal"
FC_DATETIME_LESS = "datetime_less"
FC_DATETIME_LESS_THAN_OR_EQUAL = "datetime_less_than_or_equal"

import sys
if sys.version_info[:3] > (3, 8, 0):
    from typing import TypedDict, Literal


    BooleanFilterCondition = Literal[
        'boolean_is_true',
        'boolean_is_false'
    ]
    StringFilterCondition = Literal[
        'contains',
        'string_does_not_contain',
        'string_exactly',
        'string_not_exactly',
        'string_starts_with',
        'string_ends_with'
    ]
    NumberFilterCondition = Literal[
        'number_exactly',
        'number_not_exactly',
        'greater',
        'greater_than_or_equal',
        'less',
        'less_than_or_equal',
        'number_lowest',
        'number_highest',
    ]
    DatetimeFilterCondition = Literal[
        'datetime_exactly',
        'datetime_not_exactly',
        'datetime_greater',
        'datetime_greater_than_or_equal',
        'datetime_less',
        'datetime_less_than_or_equal'
    ]
    SharedFilterCondition = Literal[
        'empty',
        'not_empty',
        'most_frequent',
        'least_frequent'
    ]
    OperatorType = Literal[
        'And', 
        'Or'
    ]
    
    PrimitiveTypeName = Literal[
        'str', 
        'int', 
        'float', 
        'number',
        'bool', 
        'datetime', 
        'timedelta'
    ]

    class Filter(TypedDict):
        condition: Union[BooleanFilterCondition, StringFilterCondition, NumberFilterCondition, DatetimeFilterCondition, SharedFilterCondition]
        value: Union[str, float, int]

    class FilterGroup(TypedDict):
        # NOTE: this is a recursive type. The filter group can contain a filter group
        filters: List[Union[Filter, "FilterGroup"]] #type: ignore
        operator: OperatorType #type:ignore

    class ColumnIDWithFilter(TypedDict):
        column_id: ColumnID
        filter: Filter

    class ColumnIDWithFilterGroup(TypedDict):
        column_id: ColumnID
        filter: FilterGroup

    class ColumnHeaderWithFilter(TypedDict):
        column_header: ColumnHeader
        filter: Filter

    class ColumnIDWithPivotTransform(TypedDict):
        column_id: ColumnID
        transformation: PivotColumnTransformation

    class ColumnHeaderWithPivotTransform(TypedDict):
        column_header: ColumnHeader
        transformation: PivotColumnTransformation

    class ExcelRangeRangeImport(TypedDict):
        type: Literal['range']
        df_name: str
        value: str

    class ExcelRangeDynamicImport(TypedDict):
        type: Literal['dynamic']
        df_name: str
        start_condition: Dict[str, Any]
        end_condition: Dict[str, Any]
        column_end_condition: Dict[str, Any]

    ExcelRangeImport = Union[ExcelRangeRangeImport, ExcelRangeDynamicImport]

    class CodeSnippet(TypedDict):
        Id: str
        Name: str
        Description: str
        Code: List[str]


    class SnowflakeCredentials(TypedDict):
        type: str
        username: str
        password: str 
        account: str

    class SnowflakeTableLocationAndWarehouseOptional(TypedDict):
        role: Optional[str]
        warehouse: Optional[str] 
        database: Optional[str]
        schema: Optional[str]
        table_or_view: Optional[str] 

    class SnowflakeTableLocationAndWarehouse(TypedDict):
        role: str
        warehouse: str
        database: str
        schema: str
        table_or_view: str

    class SnowflakeQueryParams(TypedDict):
        columns: List[str]
        limit: Optional[int]
        
    class CodeSnippetEnvVars(TypedDict):
        MITO_CONFIG_CODE_SNIPPETS_VERSION: str
        MITO_CONFIG_CODE_SNIPPETS_URL: str
        MITO_CONFIG_CODE_SNIPPETS_SUPPORT_EMAIL: Optional[str]

    class RawParserMatch(TypedDict):
        type: Literal['{HEADER}', '{INDEX}', '{SHEET}']
        substring_range: ParserMatchSubstringRange
        unparsed: str
        parsed: Any
        row_offset: RowOffset

    class ParserMatch(TypedDict):
        type: Literal['{HEADER}', '{HEADER}{INDEX}', '{HEADER}:{HEADER}', '{HEADER}{INDEX}:{HEADER}{INDEX}', '{SHEET}!{HEADER}:{HEADER}']
        substring_range: ParserMatchSubstringRange
        unparsed: str
        parsed: Any
        row_offset: Union[RowOffset, Tuple[RowOffset, RowOffset]]

    class FrontendFormulaString(TypedDict):
        type: Literal['string part']
        string: str

    class FrontendFormulaHeaderIndexReference(TypedDict):
        type: Literal['{HEADER}{INDEX}']
        display_column_header: str
        row_offset: Optional[int]

    class FrontendFormulaHeaderReference(TypedDict):
        type: Literal['{HEADER}']
        display_column_header: str

    class FrontendFormulaSheetReference(TypedDict):
        type: Literal['{SHEET}']
        display_sheet_name: str

    class FormulaLocationEntireColumn(TypedDict):
        type: Literal['entire_column']

    class FormulaLocationToSpecificIndexLabels(TypedDict):
        type: Literal['specific_index_labels']
        index_labels: List[Any]

    class Selection(TypedDict):
        selected_df_name: str
        selected_column_headers: List[ColumnHeader]
        selected_row_labels: List[IndexLabel]

    class DataframeReconData(TypedDict):
        created_dataframes: Dict[str, pd.DataFrame]
        deleted_dataframes: List[str]
        modified_dataframes: Dict[str, pd.DataFrame]
        last_line_expression_value: Optional[Any]
        prints: str

    class ColumnReconData(TypedDict):
        created_columns: List[ColumnHeader]
        deleted_columns: List[ColumnHeader]
        modified_columns: List[ColumnHeader]
        renamed_columns: Dict[ColumnHeader, ColumnHeader]

    class ModifiedDataframeReconData(TypedDict):
        column_recon: ColumnReconData
        num_added_or_removed_rows: int

    class AITransformFrontendResult(TypedDict):
        last_line_value: Optional[Union[str, bool, int, float, np.number]]
        created_dataframe_names: List[str]
        deleted_dataframe_names: List[str]
        modified_dataframes_recons: Dict[str, ModifiedDataframeReconData]
        prints: str

    ParamName = str
    # NOTE: these cannot be changed, as they are part of the public interface, They are exposed through code-options
    # function param specification - where you can pass the param subtype, to automatically generate params for all 
    # of that subtype
    ParamType = Literal[
        'import',
        'export'
    ]
    ParamSubtype = Literal[
        'import_dataframe',
        'file_name_export_excel',
        'file_name_export_csv',
        'file_name_import_excel',
        'file_name_import_csv',
        'all' # This represents all of the above
    ]
    ParamValue = str

    # For streamlit applications, when we want to display the parameters to the user, 
    # we need to know various metadata about the parameter
    class ParamMetadata(TypedDict):
        type: ParamType
        subtype: ParamSubtype
        required: bool
        name: str
        original_value: Optional[str]

    # You can either pass in: 'all', which will generate all the params for the given subtype
    # Or you can pass in a list of the param subtypes you want to parameterize
    # Or a dictionary that maps the starting param value to the new param name, which will 
    # parameterize that specific param
    CodeOptionsFunctionParams = Union[OrderedDict, ParamSubtype, List[ParamSubtype]]

    class CodeOptions(TypedDict):
        as_function: bool
        call_function: bool
        function_name: str
        function_params: CodeOptionsFunctionParams # type: ignore

        # The params below become optional. Typing them is hard, so use care when accessing them
        import_custom_python_code: bool

    UserDefinedFunctionParamType = Literal['any', 'str', 'int', 'float', 'bool', 'DataFrame', 'ColumnHeader']

    class MitoTheme(TypedDict):
        primaryColor: str
        backgroundColor: str
        secondaryBackgroundColor: str
        textColor: str

    class MitoFrontendSelection(TypedDict):
        startingRowIndex: int
        endingRowIndex: int
        startingColumnIndex: int
        endingColumnIndex: int

    class MitoFrontendIndexAndSelections(TypedDict):
        selectedDataframeIndex: int
        selections: List[MitoFrontendSelection]

    class OverwriteSheetIndexParams(TypedDict):
        sheet_index_to_overwrite: int
        attempt_to_save_filter_metadata: bool

    class ExecuteThroughTranspileNewDataframeParams(TypedDict):
        new_df_names: List[str]
        df_source: str
        overwrite: Optional[OverwriteSheetIndexParams]

else:
    Filter = Any #type: ignore
    FilterGroup = Any #type: ignore
    OperatorType = Any #type:ignore
    PrimitiveTypeName = None # type: ignore
    ColumnIDWithFilter = Any # type:ignore
    ColumnIDWithFilterGroup = Any # type:ignore
    ColumnHeaderWithFilter = Any # type:ignore
    ColumnIDWithPivotTransform = Any # type:ignore
    ColumnHeaderWithPivotTransform = Any # type:ignore
    ExcelRangeImport = Any # type:ignore
    CodeSnippet = Any # type:ignore
    CodeSnippetEnvVars = Any # type:ignore
    RawParserMatch = Any # type:ignore
    ParserMatch = Any # type:ignore
    SnowflakeCredentials = Any # type:ignore
    SnowflakeTableLocationAndWarehouse = Any # type:ignore
    SnowflakeTableLocationAndWarehouseOptional = Any #type:ignore
    SnowflakeQueryParams = Any # type:ignore
    CodeSnippetEnvVars = Any # type:ignore
    FrontendFormulaString = Any # type:ignore
    FrontendFormulaHeaderIndexReference = Any # type:ignore
    FrontendFormulaHeaderReference = Any # type:ignore
    FrontendFormulaSheetReference = Any # type:ignore
    FormulaLocationEntireColumn = Any # type:ignore
    FormulaLocationToSpecificIndexLabels = Any # type:ignore
    Selection = Any # type:ignore
    DataframeReconData = Any # type: ignore
    ColumnReconData = Any # type: ignore
    ModifiedDataframeReconData = Any # type: ignore
    AITransformFrontendResult = Any # type: ignore
    CodeOptions = Any # type: ignore
    UserDefinedImporterParamType = Any # type: ignore
    OverwriteSheetIndexParams = Any # type: ignore
    ExecuteThroughTranspileNewDataframeParams = Any # type: ignore
    UserDefinedFunctionParamType = Any # type: ignore
    MitoTheme = Any # type: ignore
    MitoFrontendSelection = Any # type: ignore
    MitoFrontendIndexAndSelections = Any # type: ignore

    ParamName = str # type: ignore
    ParamType = str # type: ignore
    ParamSubtype = str # type: ignore
    ParamValue = str # type: ignore
    ParamMetadata = Any # type: ignore

    CodeOptionsFunctionParams = Any # type: ignore


FrontendFormulaPart = Union[FrontendFormulaString, FrontendFormulaHeaderIndexReference, FrontendFormulaHeaderReference, FrontendFormulaSheetReference]
FrontendFormula = List[FrontendFormulaPart]

FormulaAppliedToType = Union[FormulaLocationEntireColumn, FormulaLocationToSpecificIndexLabels]


if sys.version_info[:3] > (3, 8, 0):
    from typing import TypedDict

    class FrontendFormulaAndLocation(TypedDict):
        frontend_formula: FrontendFormula
        location: FormulaAppliedToType
        index: List[Any]

else:
    FrontendFormulaAndLocation = Any # type:ignore


#!/usr/bin/env python
# coding: utf-8

# Copyright (c) Mito.
# Distributed under the terms of the Modified BSD License.
from io import StringIO
import json
import os
import pandas as pd
from mitosheet.transpiler.transpile_utils import get_default_code_options
import pytest
import numpy as np

from mitosheet.mito_backend import MitoBackend, get_mito_backend
from mitosheet.tests.test_utils import create_mito_wrapper_with_data, create_mito_wrapper
from mitosheet.transpiler.transpile import transpile
from mitosheet.tests.decorators import pandas_post_1_only
from mitosheet.utils import MAX_COLUMNS


def test_example_creation_blank():
    df = pd.DataFrame()
    w = MitoBackend(df)

VALID_DATAFRAMES = [
    (pd.DataFrame(), None),
    (pd.DataFrame(data={'A': [1, 2, 3]}), None),
    (pd.DataFrame(data={'A0123': [1, 2, 3]}), None),
    (pd.DataFrame(data={0: [1, 2, 3]}), None),
    (pd.DataFrame(data={0.1: [1, 2, 3]}), None),
    (pd.DataFrame(data={'A A A A': [1, 2, 3], 0: [1, 2, 3]}), None),
    (pd.DataFrame(data={'         a         ': [1, 2, 3], '123': [1, 2, 3]}), None),
    (pd.DataFrame(data={'## this is a test': [1, 2, 3], ' !': [1, 2, 3]}), None),
    (pd.DataFrame(data={'TOTAL': [1, 2, 3], '#123': [1, 2, 3]}), None),
    (pd.DataFrame(data={'NUMber': [1, 2, 3], '.,,': [1, 2, 3]}), None),
    (pd.DataFrame(data={'this is a possible ! column header that could be there': [1, 2, 3], '.,,': [1, 2, 3]}), None),
    (pd.DataFrame(data={1000.123123: [1, 2, 3], 52.100: [1, 2, 3]}), None),

    # to_csv of file content
    (pd.DataFrame(data={'A': [1, 2, 3], 'B': [2, 3, 4]}).to_csv(index=False), pd.DataFrame(data={'A': [1, 2, 3], 'B': [2, 3, 4]})),
    # to_json of file content
    (pd.DataFrame(data={'A': [1, 2, 3], 'B': [2, 3, 4]}).to_json(), pd.DataFrame(data={'A': [1, 2, 3], 'B': [2, 3, 4]})),
    # to_dict of file content
    (pd.DataFrame(data={'A': [1, 2, 3], 'B': [2, 3, 4]}).to_dict('records'), pd.DataFrame(data={'A': [1, 2, 3], 'B': [2, 3, 4]})),
    # a list of strings (which are to_csv of file content)
    ([pd.DataFrame(data={'A': [1, 2, 3], 'B': [2, 3, 4]}).to_csv(index=False), pd.DataFrame(data={'A': [1, 2, 3], 'B': [2, 3, 4]}).to_csv(index=False)], pd.DataFrame(data={'A': [1, 2, 3], 'B': [2, 3, 4]})),
]
@pytest.mark.parametrize("df", VALID_DATAFRAMES)
def test_df_creates_valid_df(df):
    print(type(df[0]))
    mito = create_mito_wrapper(df[0])
    assert mito.mito_backend is not None

    if df[1] is None:
        assert mito.mito_backend.steps_manager.curr_step.dfs[0].equals(df[0])
    else:
        assert mito.mito_backend.steps_manager.curr_step.dfs[0].equals(df[1])

def test_df_with_nan_creates_backend():
    df = pd.DataFrame(data={np.nan: [1, 2, 3], 52.100: [1, 2, 3]})
    mito = get_mito_backend(df)
    assert mito is not None
    assert np.array_equal(mito.steps_manager.curr_step.dfs[0].keys(), list(df.keys()), equal_nan=True)


def test_create_with_multiple_dataframes():
    mito = get_mito_backend(pd.DataFrame(), pd.DataFrame(), pd.DataFrame(), pd.DataFrame(data={'A': [1, 2, 3]}))
    assert mito is not None

def test_can_call_sheet_with_just_filename():
    df = pd.DataFrame(data={'A': [1, 2, 3], 'B': [2, 3, 4]})
    df.to_csv('../1.csv', index=False)

    mito = get_mito_backend('../1.csv')

    assert len(mito.steps_manager.dfs) == 1
    assert mito.steps_manager.dfs[0].equals(df)

    # Remove the test file
    os.remove('../1.csv')


def test_can_call_sheet_with_filename_mulitiple_times():
    df = pd.DataFrame(data={'A': [1, 2, 3], 'B': [2, 3, 4]})
    df.to_csv('../1.csv', index=False)

    mito = get_mito_backend('../1.csv', '../1.csv')

    assert len(mito.steps_manager.dfs) == 2
    assert mito.steps_manager.dfs[0].equals(df)
    assert mito.steps_manager.dfs[1].equals(df)
    assert mito.steps_manager.curr_step.df_names[0] != mito.steps_manager.curr_step.df_names[1]

    # Remove the test file
    os.remove('../1.csv')


def test_can_call_sheet_with_df_and_filename():
    df = pd.DataFrame(data={'A': [1, 2, 3], 'B': [2, 3, 4]})
    df.to_csv('../1.csv', index=False)

    mito = get_mito_backend(df, '../1.csv')

    assert len(mito.steps_manager.dfs) == 2
    assert mito.steps_manager.dfs[0].equals(mito.steps_manager.dfs[1])

    # Remove the test file
    os.remove('../1.csv')

    code_container = transpile(mito.steps_manager)

    assert code_container == [
        'from mitosheet.public.v3 import *',
        'import pandas as pd',
        '',
        '# Read in filepaths as dataframes',
        'df_1 = pd.read_csv(r\'../1.csv\')',
        '',
    ]

def test_can_use_utf_16_when_passing_string():
    df = pd.DataFrame(data={'A': [1, 2, 3], 'B': [2, 3, 4]})

    encoding = 'UTF-16'
    df.to_csv('test_file_path.csv', index=False, encoding=encoding)

    # Create with no dataframes
    mito = get_mito_backend('test_file_path.csv')
    # And then import just a test file

    assert mito.steps_manager.dfs[0].equals(df)

    code_container = transpile(mito.steps_manager)

    assert code_container == [
        'from mitosheet.public.v3 import *',
        'import pandas as pd',
        '',
        '# Read in filepaths as dataframes',
        f'test_file_path = pd.read_csv(r\'test_file_path.csv\', encoding=\'{encoding}\')',
        '',
    ]
    os.remove('test_file_path.csv')


def test_call_makes_copies():
    df = pd.DataFrame(data={'A A': [1, 2, 3]})
    get_mito_backend(df)

    # Test we don't change the headers!
    assert df.columns.tolist() == ['A A']

def test_can_call_with_indexes():
    df = pd.DataFrame(data={'A': [1, 2, 3], 'B': ['A', 'B', 'C'], 'D': ['E', 'F', 'G']})

    string_index = df.set_index('B')
    get_mito_backend(string_index)

    multi_index = df.set_index(['B', 'D'])
    get_mito_backend(multi_index)

def test_sheet_json_holds_all_columns():
    df = pd.DataFrame({i: [1, 2, 3] for i in range(MAX_COLUMNS + 100)})
    mito = create_mito_wrapper(df)
    sheet_data = json.loads(mito.sheet_data_json)[0]
    for i in range(MAX_COLUMNS, MAX_COLUMNS + 100):
        assert sheet_data['columnIDsMap'][str(i)] is not None
        assert sheet_data['columnFormulasMap'][str(i)] is not None
        assert sheet_data['columnFiltersMap'][str(i)] is not None
        assert sheet_data['columnIDsMap'][str(i)] is not None
        assert sheet_data['columnDtypeMap'][str(i)] is not None

@pandas_post_1_only
def test_throws_duplicated_column_error():
    with pytest.raises(ValueError) as e_info:
        df = pd.DataFrame(columns=['A', 'A'])
        get_mito_backend(df)
    assert 'A' in str(e_info)

def test_create_mito_backend_with_string_names_set_df_names():
    df = pd.DataFrame({i: [1, 2, 3] for i in range(MAX_COLUMNS + 100)})
    df.to_csv('test_file.csv', index=False)
    mito = create_mito_wrapper('test_file.csv')
    assert mito.mito_backend.steps_manager.curr_step.final_defined_state.df_names == ['test_file']
    os.remove('test_file.csv')

def test_create_backend_with_code_options_works():
    code_options = get_default_code_options('tmp')
    code_options['call_function'] = False
    mito_backend = MitoBackend(code_options=code_options)
    assert mito_backend.steps_manager.code_options == code_options


#!/usr/bin/env python
# coding: utf-8

# Copyright (c) Saga Inc.
# Distributed under the terms of the GPL License.
from datetime import datetime
from typing import Any, Dict

from mitosheet.telemetry.telemetry_utils import identify
from mitosheet.types import StepsManagerType
from mitosheet.user import get_user_field, set_user_field
from mitosheet.user.schemas import UJ_FEEDBACKS_V2
    

UPDATE_FEEDBACK_V2_OBJECT_EVENT = 'update_feedback_v2_obj_update'
UPDATE_FEEDBACK_V2_OBJECT_EVENT_PARAMS = ['feedback_id', 'num_usages', 'questions_and_answers']

def execute_add_feedback_update(steps_manager: StepsManagerType, feedback_id: str, num_usages: int, questions_and_answers: Dict[str, Any]) -> None:
    """
    The function responsible for adding the feedback to the feedback_v2 object
    in the user.json. 
    """

    user_feedback_obj = get_user_field(UJ_FEEDBACKS_V2)
    if user_feedback_obj is None: 
        user_feedback_obj = {}

    if feedback_id not in user_feedback_obj:
        user_feedback_obj[feedback_id] = []

    # Create the feedback object for the feedback we're adding
    # to the user.json
    new_feedback_obj = {
        'date': datetime.today().strftime('%Y-%m-%d'),
        'num_usages': num_usages,
        'questions_and_answers': questions_and_answers
    }
    
    # Add the new feedback to the feedback_obj
    user_feedback_obj[feedback_id].append(new_feedback_obj)
    
    # Set the field to the new list
    set_user_field(UJ_FEEDBACKS_V2, user_feedback_obj)

    # Identify just in case
    identify()


UPDATE_FEEDBACK_V2_OBJECT_UPDATE = {
    'event_type': UPDATE_FEEDBACK_V2_OBJECT_EVENT,
    'params': UPDATE_FEEDBACK_V2_OBJECT_EVENT_PARAMS,
    'execute': execute_add_feedback_update
}



#!/usr/bin/env python
# coding: utf-8

# Copyright (c) Saga Inc.
# Distributed under the terms of the GPL License.
"""
Contains tests to make sure that the mito analytics test is
performing correctly
"""

import pprint
import sys
import time
import pandas as pd
from mitosheet.column_headers import get_column_header_id
from mitosheet.enterprise.mito_config import MITO_CONFIG_LOG_SERVER_BATCH_INTERVAL, MITO_CONFIG_LOG_SERVER_URL, MITO_CONFIG_VERSION
from mitosheet.errors import MitoError
from mitosheet.telemetry.telemetry_utils import PRINT_LOGS
import pytest
from unittest.mock import patch
import os
import json
from mitosheet.tests.test_mito_config import delete_all_mito_config_environment_variables
from mitosheet.tests.test_utils import create_mito_wrapper, create_mito_wrapper_with_data
from mitosheet.types import FORMULA_ENTIRE_COLUMN_TYPE
from mitosheet.utils import get_new_id

try:
    from pandas import __version__ as pandas_version
except:
    pandas_version = 'no pandas'
try:
    # Format version_python as "3.8.5"
    version_python = '.'.join([str(x) for x in sys.version_info[:3]])
except:
    version_python = 'no python'

URL = "https://url?" 

def test_not_printing_logs():
    assert PRINT_LOGS is False

def test_log_uploader_single_edit_event():
    
    os.environ[MITO_CONFIG_VERSION] = "2"
    os.environ[MITO_CONFIG_LOG_SERVER_URL] =  f"{URL}"
    os.environ[MITO_CONFIG_LOG_SERVER_BATCH_INTERVAL] = "0"
    
    mito = create_mito_wrapper_with_data([123])

    with patch('requests.post') as mock_post:
        mito.add_column(0, 'B')

        print(mock_post.call_args_list)
        log_call = mock_post.call_args_list[0]

        # Get the URL from the log call
        actual_url = log_call[0][0]
        assert actual_url == URL

        # From the call object, get the payload 
        # that was passed to requests.post
        data = log_call[1]['data']
        log_payload = json.loads(data)[0]

        assert len(log_payload) == 9
        assert log_payload["params_sheet_index"] == 0
        assert log_payload["params_column_header"] is not None
        assert log_payload["params_column_header_index"] == -1
        assert log_payload["params_public_interface_version"] == 3
        assert log_payload["version_python"] == version_python
        assert log_payload["version_pandas"] == pandas_version
        assert log_payload["version_mito"] is not None
        assert log_payload["timestamp_gmt"] is not None
        assert log_payload["event"] == "add_column_edit"

    delete_all_mito_config_environment_variables()

def test_log_uploader_multiple_edit_events():
    
    os.environ[MITO_CONFIG_VERSION] = "2"
    os.environ[MITO_CONFIG_LOG_SERVER_URL] =  f"{URL}"
    os.environ[MITO_CONFIG_LOG_SERVER_BATCH_INTERVAL] = "1"
    
    mito = create_mito_wrapper_with_data([123])

    mito.add_column(0, 'B')
    mito.delete_columns(0, ['B'])

    # wait 1 second
    time.sleep(1)

    with patch('requests.post') as mock_post:

        mito.add_column(0, 'C')

        # Even though multiple events occured, there should be only one upload
        assert len(mock_post.call_args_list) == 1

        log_call = mock_post.call_args_list[0]

        # Get the URL from the log call
        actual_url = log_call[0][0]
        assert actual_url == URL

        # From the call object, get the payload 
        # that was passed to requests.post
        data = log_call[1]['data']
        add_column_log_event = json.loads(data)[0]

        assert len(add_column_log_event) == 9
        assert add_column_log_event["params_sheet_index"] == 0
        assert add_column_log_event["params_column_header"] is not None
        assert add_column_log_event["params_column_header_index"] == -1
        assert add_column_log_event["params_public_interface_version"] == 3
        assert add_column_log_event["version_python"] == version_python
        assert add_column_log_event["version_pandas"] == pandas_version
        assert add_column_log_event["version_mito"] is not None
        assert add_column_log_event["timestamp_gmt"] is not None
        assert add_column_log_event["event"] == "add_column_edit"

        delete_columns_log_event = json.loads(data)[1]
        print(delete_columns_log_event)
        assert len(delete_columns_log_event) == 8
        assert delete_columns_log_event["params_sheet_index"] == 0
        assert len(delete_columns_log_event["params_column_ids"]) == 1
        assert delete_columns_log_event["params_public_interface_version"] == 3
        assert delete_columns_log_event["version_python"] == version_python
        assert delete_columns_log_event["version_pandas"] == pandas_version
        assert delete_columns_log_event["version_mito"] is not None
        assert delete_columns_log_event["timestamp_gmt"] is not None
        assert delete_columns_log_event["event"] == "delete_column_edit"

    delete_all_mito_config_environment_variables()

def test_log_uploader_error_events():
    
    os.environ[MITO_CONFIG_VERSION] = "2"
    os.environ[MITO_CONFIG_LOG_SERVER_URL] =  f"{URL}"
    os.environ[MITO_CONFIG_LOG_SERVER_BATCH_INTERVAL] = "0"
    
    mito = create_mito_wrapper(pd.DataFrame({'A': [1, 2, 3]}))

    with patch('requests.post') as mock_post:

        mito.set_formula('=INVALID_FUNCTION()', 0, 'A')

        assert len(mock_post.call_args_list) == 1

        log_call = mock_post.call_args_list[0]

        # Get the URL from the log call
        actual_url = log_call[0][0]
        assert actual_url == URL

        # From the call object, get the payload 
        # that was passed to requests.post
        data = log_call[1]['data']
        log_event = json.loads(data)[0]

        assert len(log_event) == 13
        assert log_event["error_traceback"] is not None
        assert log_event["error_traceback_last_line"] is not None
        assert log_event["params_sheet_index"] == 0
        assert log_event["params_column_id"] is not None
        assert log_event["params_formula_label"] is not None
        assert log_event["params_index_labels_formula_is_applied_to"] is not None
        assert log_event["params_public_interface_version"] == 3
        assert log_event["params_new_formula"] == '=INVALID_FUNCTION()'
        assert log_event["version_python"] == version_python
        assert log_event["version_pandas"] == pandas_version
        assert log_event["version_mito"] is not None
        assert log_event["timestamp_gmt"] is not None
        assert log_event["event"] == "error"

    delete_all_mito_config_environment_variables()
    

def test_log_uploader_mitosheet_rendered():
    
    os.environ[MITO_CONFIG_VERSION] = "2"
    os.environ[MITO_CONFIG_LOG_SERVER_URL] = f"{URL}"
    os.environ[MITO_CONFIG_LOG_SERVER_BATCH_INTERVAL] = "0"
    
    mito = create_mito_wrapper(pd.DataFrame({'A': [1, 2, 3]}))

    with patch('requests.post') as mock_post:
        mito.mito_backend.receive_message({
            'params': {'user_agent': 'Mozilla/5.0 (Macintosh; Intel Mac OS X 10_15_7) AppleWebKit/605.1.15 (KHTML, like Gecko) Version/17.2.1 Safari/605.1.15'}, 
            'event': 'log_event', 
            'type': 'mitosheet_rendered', 
            'id': '_pt4wgfw0z'
        })

        assert len(mock_post.call_args_list) == 1

        log_call = mock_post.call_args_list[0]

        # Get the URL from the log call
        actual_url = log_call[0][0]
        assert actual_url == URL

        # From the call object, get the payload 
        # that was passed to requests.post
        data = log_call[1]['data']
        log_event = json.loads(data)[0]

        assert len(log_event) == 6
        assert log_event["params_user_agent"] == "Mozilla/5.0 (Macintosh; Intel Mac OS X 10_15_7) AppleWebKit/605.1.15 (KHTML, like Gecko) Version/17.2.1 Safari/605.1.15"
        assert log_event["version_python"] == version_python
        assert log_event["version_pandas"] == pandas_version
        assert log_event["version_mito"] is not None
        assert log_event["timestamp_gmt"] is not None
        assert log_event["event"] == "mitosheet_rendered"

    delete_all_mito_config_environment_variables()



def test_log_uploader_long_interval_does_not_trigger_upload():
    
    os.environ[MITO_CONFIG_VERSION] = "2"
    os.environ[MITO_CONFIG_LOG_SERVER_URL] =  f"{URL}"
    os.environ[MITO_CONFIG_LOG_SERVER_BATCH_INTERVAL] = "100"
    
    mito = create_mito_wrapper(pd.DataFrame({'A': [1, 2, 3]}))

    with patch('requests.post') as mock_post:

        mito.add_column(0, 'B')
        mito.add_column(0, 'C')
        mito.add_column(0, 'D')

        assert len(mock_post.call_args_list) == 0

    delete_all_mito_config_environment_variables()



#!/usr/bin/env python
# coding: utf-8

# Copyright (c) Saga Inc.
# Distributed under the terms of the GPL License.

import json
from typing import Any, Dict, List, Optional, Tuple, Union
from mitosheet.api.get_validate_snowflake_credentials import get_cached_snowflake_credentials
from mitosheet.types import MitoSafeSnowflakeConnection, SnowflakeTableLocationAndWarehouseOptional, StepsManagerType

# The snowflake-connector-python package is only available in Python > 3.6 
# and is not distributed with the mitosheet package, so we make sure to 
# note assume that the import will succeed. 
try:
        import snowflake.connector
        SNOWFLAKE_CONNECTOR_IMPORTED = True
except ImportError:
        SNOWFLAKE_CONNECTOR_IMPORTED = False


def get_snowflake_connection_or_exception(username: str, password: str, account: str) -> Union[MitoSafeSnowflakeConnection, Exception]:
        try:
                return snowflake.connector.connect(
                        user=username,
                        password=password,
                        account=account,
                )
        except Exception as e:
                return e

def get_available_snowflake_options_and_defaults(params: Dict[str, Any], steps_manager: StepsManagerType) -> Dict[str, Any]:

        if not SNOWFLAKE_CONNECTOR_IMPORTED: 
                return {
                        'type': 'error',    
                        'error_message': 'The package snowflake-connector-python is required to use this feature, but it is not accessible. Ensure it is installed.'
                }

        credentials = get_cached_snowflake_credentials()
        table_loc_and_warehouse: SnowflakeTableLocationAndWarehouseOptional = params['table_loc_and_warehouse']

        if credentials is None:
                return {
                        'type': 'error',    
                        'error_message': 'Invalid authentication information. Please try again.'
                }
                
        username = credentials['username']
        password = credentials['password']
        account = credentials['account']

        con_or_exception = get_snowflake_connection_or_exception(username, password, account)

        if isinstance(con_or_exception, Exception):
                exception = con_or_exception
                return {
                        'type': 'error',    
                        'error_message': f'{exception}'
                }

        con = con_or_exception

        _role = table_loc_and_warehouse.get('role')
        _warehouse = table_loc_and_warehouse.get('warehouse')
        _database = table_loc_and_warehouse.get('database') 
        _schema = table_loc_and_warehouse.get('schema')
        _table_or_view = table_loc_and_warehouse.get('table_or_view')

        roles = get_roles(con)
        role = _role if _role is not None else get_default_role(roles)

        # After getting the role, use it so that we get the correct list of 
        # databases, schemas, tables, etc.
        if con is not None:
                con.cursor().execute(f'USE ROLE {role}')

        warehouses = get_warehouses(con)
        warehouse = _warehouse if _warehouse is not None else get_default_warehouse(warehouses)

        databases = get_databases(con)
        database = _database if _database is not None else get_default_database(databases)

        schemas = get_schemas(con, database)
        schema = _schema if _schema is not None else get_default_schema(schemas)

        tables = get_tables(con, database, schema)
        views = get_views(con, database, schema)
        tables_and_views = tables + views
        table_or_view = _table_or_view if _table_or_view is not None else get_default_table(tables_and_views)

        columns = get_columns(con, database, schema, table_or_view)

        return {
                'type': 'success',    
                'config_options': {
                        'roles': roles,
                        'warehouses': warehouses,    
                        'databases': databases,    
                        'schemas': schemas,    
                        'tables_and_views': tables_and_views,
                        'columns': columns
                },
                'default_values': {
                        'role': role,
                        'warehouse': warehouse,
                        'database': database,
                        'schema': schema,
                        'table_or_view': table_or_view,
                },
        }

def get_roles(con: MitoSafeSnowflakeConnection) -> List[str]:
        if con is None:
                return []

        # We use SHOW GRANTS to show all of the roles granted to the user. 
        # Instead of SHOW ROLES which shows all of the roles in the account.
        cur = con.cursor().execute('SHOW GRANTS')   

        if cur is None:
                return []
        
        roles = cur.fetchall() # type: ignore
        return [role[1] for role in roles] # type: ignore

def get_warehouses(con: MitoSafeSnowflakeConnection) -> List[str]:
        if con is None: 
                return []

        # List all of the warehouses available to the user
        cur = con.cursor().execute('SHOW WAREHOUSES')

        if cur is None:
                return []

        warehouses = cur.fetchall() # type: ignore
        return [wh[0] for wh in warehouses] # type: ignore

def get_databases(con: MitoSafeSnowflakeConnection) -> List[str]:
        if con is None:
                return []

        # List all of the databases available to the user
        cur = con.cursor().execute('SHOW DATABASES')

        if cur is None:
                return []

        databases = cur.fetchall() # type: ignore
        return [db[1] for db in databases] # type: ignore

def get_schemas(con: MitoSafeSnowflakeConnection, database: Optional[str]) -> List[str]:
        if con is None: 
                return []

        # List all of the schemas in a particular database available to the user
        cur = con.cursor().execute(f'SHOW SCHEMAS in {database}')
        
        if cur is None:
                return []

        schemas = cur.fetchall() # type: ignore
        return [s[1] for s in schemas] # type: ignore

def get_tables(con: MitoSafeSnowflakeConnection, database: Optional[str], schema: Optional[str]) -> List[str]:
        if con is None or database is None or schema is None:
                return []

        # List all of the tables in a schema
        cur = con.cursor().execute(f'SHOW TABLES in {database}.{schema}')

        if cur is None:
                return []

        tables = cur.fetchall() # type: ignore
        return [table[1] for table in tables] # type: ignore

def get_views(con: MitoSafeSnowflakeConnection, database: Optional[str], schema: Optional[str]) -> List[str]:
        if con is None or database is None or schema is None:
                return []

        # List all of the tables in a schema
        cur = con.cursor().execute(f'SHOW VIEWS in {database}.{schema}')

        if cur is None:
                return []

        views = cur.fetchall() # type: ignore
        return [view[1] for view in views] # type: ignore

def get_columns(con: MitoSafeSnowflakeConnection, database: Optional[str], schema: Optional[str], table: Optional[str]) -> List[str]:
        if con is None or database is None or schema is None or table is None:
                return []

        # List all of the columns in a table 
        cur = con.cursor().execute(f'SHOW COLUMNS in {database}.{schema}.{table}')

        if cur is None:
                return []

        columns = cur.fetchall() # type: ignore
        return [column[2] for column in columns] # type: ignore

def get_default_role(roles: List[str]) -> Optional[str]:
        return roles[0] if len(roles) > 0 else None

def get_default_warehouse(warehouses: List[str]) -> Optional[str]:
        return warehouses[0] if len(warehouses) > 0 else None 

def get_default_database(databases: List[str]) -> Optional[str]:
        return databases[0] if len(databases) > 0 else None 

def get_default_schema(schemas: List[str]) -> Optional[str]:
        return schemas[0] if len(schemas) > 0 else None 

def get_default_table(tables: List[str]) -> Optional[str]:
        return tables[0] if len(tables) > 0 else None




#!/usr/bin/env python
# coding: utf-8

# Copyright (c) Saga Inc.
# Distributed under the terms of the GPL License.

"""
NOTE: DO NOT CHANGE THIS FILE. THESE ARE DEPRECATED UTILS
THAT CANNOT CHANGE AND SO SHOULD NEVER BE CHANGED.
"""

import numbers
import re
from typing import Any, Callable, Dict, List
from mitosheet.telemetry.telemetry_utils import log

import pandas as pd

import warnings
import functools

from mitosheet.types import ColumnHeader

def deprecated(func: Callable) -> Callable:
    """
    This is a decorator which can be used to mark functions
    as deprecated. It will result in a warning being emitted
    when the function is used. It should be used on all 
    external functions that are deprecated.

    Taken from here: https://stackoverflow.com/questions/2536307/decorators-in-the-python-standard-lib-deprecated-specifically
    """
    @functools.wraps(func)
    def new_func(*args, **kwargs):
        warnings.simplefilter('always', DeprecationWarning)  # turn off filter
        warnings.warn("Call to deprecated function {}. Support for this function will be removed in the coming months.".format(func.__name__),
                      category=DeprecationWarning,
                      stacklevel=2)
        warnings.simplefilter('default', DeprecationWarning)  # reset filter
        # Log the call
        log('used_deprecated_function', {'function_name': func.__name__})
        return func(*args, **kwargs)
    return new_func


def make_valid_header(column_header: Any) -> str:
    """
    Takes a header, and performs replaces against common characters
    to make the column_header valid!

    NOTE: DO NOT CHANGE THIS FILE. THESE ARE DEPRECATED UTILS
    THAT CANNOT CHANGE AND SO SHOULD NEVER BE CHANGED.
    """
    # If it's a tuple, we turn it into a string before continuing
    if isinstance(column_header, tuple):
        column_header = '_'.join([str(c) for c in column_header]).strip()

    # If it's just numbers, turn it into a string (with an underscore)
    if isinstance(column_header, numbers.Number):
        return str(column_header).replace('.', '_') + '_'

    # If it's just numbers in a string, add an underscore
    if column_header.isdigit():
        return column_header + "_"

    # Replace common invalid seperators with valid seperators
    replace_dict = {
        ' ': '_',
        '-': '_',
        '(': '_',
        ')': '_',
        '/': '_',
        '#': 'num',
        ',': '_',
        '.': '_',
        '!': '_',
        '?': '_'
    }
    for find, replace in replace_dict.items():
        column_header = column_header.replace(find, replace)
    
    if not is_valid_header(column_header):
        # Because we detect column headers using a word match, any word character counts
        # in a valid character
        pattern = re.compile("\w")

        new_header = ''.join([
            c for c in column_header if pattern.search(c)
        ])
        if not is_valid_header(new_header):
            # And then append an underscore, for good measure, and this should fix it!
            new_header = new_header + '_'

        return new_header
    return column_header

# We create an external version of the make_valid_header, which we
# export in 
make_valid_header_external = deprecated(make_valid_header)


def is_valid_header(column_header: ColumnHeader) -> bool:
    """
    A header is valid if it is a string that is made up of all word characters,
    with at least one non numeric char, and has at least one char.
    Valid examples: A, ABC, 012B, 213_bac, 123_123
    Invalid examples: 123, 123!!!, ABC!, 123-123
    This is a result of not being able to distingush column headers from constants
    otherwise, and would not be necessary if we had a column identifier!

    NOTE: DO NOT CHANGE THIS FILE. THESE ARE DEPRECATED UTILS
    THAT CANNOT CHANGE AND SO SHOULD NEVER BE CHANGED.
    """
    
    # Note the start and end characters in the regex, to make sure it's a full match
    return isinstance(column_header, str) and \
        len(column_header) > 0 and \
        re.compile("^\w+$").search(column_header) and \
        not column_header.isdigit() # type: ignore

def get_invalid_headers(df: pd.DataFrame) -> List[Any]:
    """
    Given a dataframe, returns a list of all the invalid headers this list has. 

    NOTE: DO NOT CHANGE THIS FILE. THESE ARE DEPRECATED UTILS
    THAT CANNOT CHANGE AND SO SHOULD NEVER BE CHANGED.
    """
    return [
        header for header in df.columns.tolist()
        if not is_valid_header(header)
    ]

def get_header_renames(column_headers: List[ColumnHeader]) -> Dict[Any, str]:
    """
    Given a list of column headers, returns a mapping from old, invalid headers to
    new, valid headers. Empty if no renames are necessary.

    NOTE: DO NOT CHANGE THIS FILE. THESE ARE DEPRECATED UTILS
    THAT CANNOT CHANGE AND SO SHOULD NEVER BE CHANGED.
    """
    renames = dict()
    for column_header in column_headers:
        if not is_valid_header(column_header):
            valid_header = make_valid_header(column_header)
            renames[column_header] = valid_header

    return renames
